#
# cvprof.rb
# $Revision: 1.3 $
#
# Copyright (c) 2004 David Holroyd
# See the file COPYING for license terms
#
# http://www.badgers-in-foil.co.uk/projects/cvprof/
#


# Library of Ruby classes used by the coverage reporter


class SourcePath
  def initialize(path)
    @source_dirs = path.split(/:/)
  end

  def locate_file(file)
    @source_dirs.each do |dir|
      filepath = "#{dir}/#{file}"
      return filepath if FileTest.exists?(filepath)
    end
    nil
  end
end


class JClassInfo
  attr_accessor :signature, :name, :source, :source_available

  def initialize(sig, source)
    unless sig =~ /^L(.*);$/
      raise "unexpected class signature format: #{sig}"
    end

    @name = $1.gsub(/\//, ".")
    @signature = sig
    @source = source
    @methods = Array.new
  end

  def package_name
    @name.sub(/\.[^\.]+$/, "")
  end

  def short_name
    @name =~ /([^\.]+)$/
    $1.sub(/\$/, '.')
  end

  def source_path
    "#{package_name.gsub(/\./, '/')}/#{@source}"
  end

  def add_method(jmethod)
    @methods << jmethod
  end

  def method_by_namesig(name_sig)
    @methods.each do |jmethod|
      return jmethod if jmethod.name_sig == name_sig
    end
    nil
  end

  def each_method
    @methods.each do |jmethod|
      yield jmethod
    end
  end
end


class JPackage
  def initialize(name)
    @name = name
    @classes = Array.new
    @sources = Hash.new
  end

  attr_accessor :name

  def add_class(jclass)
    @classes << jclass
    src = $source_path.locate_file(jclass.source_path)
    if src.nil?
      $stderr.puts "Unable to locate source file '#{jclass.source_path}' for class #{jclass.name}"
      jclass.source_available = false
    else
      jsource = @sources[src]
      if jsource.nil?
        jsource = JSource.new(jclass.source, src)
        @sources[src] = jsource
      end
      jsource.add_class(jclass)
    end
  end

  def each_source
    @sources.each_value do |src|
      yield src
    end
  end

  def each_class
    @classes.each do |src|
      yield src
    end
  end
end


class JMethod
  attr_accessor :name, :name_sig, :signature, :coverage, :bytecode, :line_info

  def initialize(name_sig)
    name_sig =~ /([^(]+)(.+)/
    @name_sig = name_sig
    @name = $1
    @signature = $2
  end
end


class JSource
  attr_accessor :name, :path
  def initialize(name, path)
    @name = name
    @path = path
    @classes = Array.new
  end

  def add_class(jclass)
    @classes << jclass
  end

  COV_FORMAT = "ia*"
  LINEINFO_FORMAT = "ia*"
  BYTECODEINFO_FORMAT = "ia*"
  LINE_FORMAT = "iii"  # "qi" in perl; ruby doesn't allow for 64 bit val

  def create_coverage_info
    @classes.each do |jclass|
      $db.each do |key, value|
        if key =~ /^#{Regexp.escape(jclass.signature)}(.*):coverage$/
          jmethod = jclass.method_by_namesig($1)
          if jmethod.nil?
            jmethod = JMethod.new($1)
            jclass.add_method(jmethod)
          end
          jmethod.coverage = BytecodeCoverageInfo.new(*value.unpack(COV_FORMAT))
        elsif key =~ /^#{Regexp.escape(jclass.signature)}(.*):lines$/
          jmethod = jclass.method_by_namesig($1)
          if jmethod.nil?
            jmethod = JMethod.new($1)
            jclass.add_method(jmethod)
          end
          (line_count, line_data) = value.unpack(LINEINFO_FORMAT)
          line_info = LineNumberingInfo.new
          line_count.times do |i|
            dat = line_data[12*i, 12*i+12]
            # TODO: endianess problem here, I expect,
            (loc_lo, loc_hi, line_no) = dat.unpack(LINE_FORMAT)
            loc = loc_hi
            loc << 32 if loc_hi != 0
            loc += loc_lo
            line_info.add(loc, line_no)
          end

          jmethod.line_info = line_info
        elsif key =~ /^#{Regexp.escape(jclass.signature)}(.*):bytecode$/
          jmethod = jclass.method_by_namesig($1)
          if jmethod.nil?
            jmethod = JMethod.new($1)
            jclass.add_method(jmethod)
          end
          (bytecode_count, bytecode_data) = value.unpack(BYTECODEINFO_FORMAT)
          jmethod.bytecode = BytecodeInfo.new(bytecode_count, bytecode_data)
        end
      end
    end
    SourceCoverageInfo.new(@classes)
  end
end


class BytecodeCoverageInfo
  def initialize(size, bitset)
    @size = size
    @bitset = bitset
  end

  def covered?(location)
    check_location(location)
    (@bitset[location/8] & (1<<(location%8))) != 0
  end

  def mark(location)
    check_location(location)
    @bitset[location/8] |= (1<<(location%8))
  end

  def size
    @size
  end

  def dump(io=$stdout)
    io.print "cv[#{@size}]:"
    @size.times do |i|
      io.print covered?(i) ? "1" : "0"
    end
    io.puts "."
  end

 private

  def check_location(location)
    raise "location less than 0: #{location}" if location < 0
    raise "location not less than #{@size}: #{location}" if location >= size
  end
end

class BytecodeInfo
  def initialize(size, code)
    @size = size
    @code = code
  end

  def each_bytecode
    counter = 0
    while counter < @size
      opcode = @code[counter]
      yield(counter, opcode_name(opcode, counter))
#puts "#{counter} #{opcode_name(opcode, counter)} + #{no_of_operands(opcode, counter)}"
      counter += 1 + no_of_operands(opcode, counter)
    end
  end

  def dump
    each_bytecode do |offset, name|
      printf("%5d %s\n", offset, name)
    end
  end

  # hack to try and allow display of bytecodes captured in 1.4 JVM
  def opcode_quick_name(opcode)
    case opcode
      when OP_GETFIELD2_QUICK then return "getfield2_quick"
    end
    return ILLEGAL_OPCODE
  end

  def opcode_name(opcode, index)
    if opcode == OP_WIDE
      return "<wide>#{PCODE_NAMES[@code[index+1]]}"
    else
      name = PCODE_NAMES[opcode]
      if name == ILLEGAL_OPCODE
        name = opcode_quick_name(opcode)
      end
      return name
    end
  end

  ILLEGAL_OPCODE = "<illegal opcode>"

  PCODE_NAMES = [
     "nop", "aconst_null", "iconst_m1", "iconst_0", "iconst_1",
     "iconst_2", "iconst_3", "iconst_4", "iconst_5", "lconst_0",
     "lconst_1", "fconst_0", "fconst_1", "fconst_2", "dconst_0",
     "dconst_1", "bipush", "sipush", "ldc", "ldc_w", "ldc2_w", "iload",
     "lload", "fload", "dload", "aload", "iload_0", "iload_1", "iload_2",
     "iload_3", "lload_0", "lload_1", "lload_2", "lload_3", "fload_0",
     "fload_1", "fload_2", "fload_3", "dload_0", "dload_1", "dload_2",
     "dload_3", "aload_0", "aload_1", "aload_2", "aload_3", "iaload",
     "laload", "faload", "daload", "aaload", "baload", "caload", "saload",
     "istore", "lstore", "fstore", "dstore", "astore", "istore_0",
     "istore_1", "istore_2", "istore_3", "lstore_0", "lstore_1",
     "lstore_2", "lstore_3", "fstore_0", "fstore_1", "fstore_2",
     "fstore_3", "dstore_0", "dstore_1", "dstore_2", "dstore_3",
     "astore_0", "astore_1", "astore_2", "astore_3", "iastore", "lastore",
     "fastore", "dastore", "aastore", "bastore", "castore", "sastore",
     "pop", "pop2", "dup", "dup_x1", "dup_x2", "dup2", "dup2_x1",
     "dup2_x2", "swap", "iadd", "ladd", "fadd", "dadd", "isub", "lsub",
     "fsub", "dsub", "imul", "lmul", "fmul", "dmul", "idiv", "ldiv",
     "fdiv", "ddiv", "irem", "lrem", "frem", "drem", "ineg", "lneg",
     "fneg", "dneg", "ishl", "lshl", "ishr", "lshr", "iushr", "lushr",
     "iand", "land", "ior", "lor", "ixor", "lxor", "iinc", "i2l", "i2f",
     "i2d", "l2i", "l2f", "l2d", "f2i", "f2l", "f2d", "d2i", "d2l", "d2f",
     "i2b", "i2c", "i2s", "lcmp", "fcmpl", "fcmpg",
     "dcmpl", "dcmpg", "ifeq", "ifne", "iflt", "ifge", "ifgt", "ifle",
     "if_icmpeq", "if_icmpne", "if_icmplt", "if_icmpge", "if_icmpgt",
     "if_icmple", "if_acmpeq", "if_acmpne", "goto", "jsr", "ret",
     "tableswitch", "lookupswitch", "ireturn", "lreturn", "freturn",
     "dreturn", "areturn", "return", "getstatic", "putstatic", "getfield",
     "putfield", "invokevirtual", "invokespecial", "invokestatic",
     "invokeinterface", ILLEGAL_OPCODE, "new", "newarray", "anewarray",
     "arraylength", "athrow", "checkcast", "instanceof", "monitorenter",
     "monitorexit", "wide", "multianewarray", "ifnull", "ifnonnull",
     "goto_w", "jsr_w", "breakpoint", ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE, ILLEGAL_OPCODE,
     ILLEGAL_OPCODE, "impdep1", "impdep2"
  ]

  OP_IINC = 132
  OP_TABLESWITCH = 170
  OP_LOOKUPSWITCH = 171
  OP_WIDE = 196

  # sun JVM's internal '_quick' opcode varients
  OP_GETFIELD2_QUICK = 208
  
  # hack to try and allow display of bytecodes captured in 1.4 JVM
  def no_of_operands_quick(opcode, index)
    case opcode
      when OP_GETFIELD2_QUICK then return 2
    end
    raise "no_of_operands undefined for opcode #{opcode}:#{opcode_name(opcode, index)}"
  end

  def no_of_operands(opcode, index)
    no = NO_OF_OPERANDS[opcode]
    case no
      when UNDEFINED then no_of_operands_quick(opcode, index)
      when RESERVED then raise "no_of_operands undefined for reserved opcode #{opcode}"
      when UNPREDICTABLE
        case opcode
          when OP_TABLESWITCH
            index += 1
            pad = index % 4
            pad = pad!=0 ? (4-pad) : 0
            index += pad
            high = get_int(index)
            low = get_int(index+4)
            length = high - low + 1
            return pad + 4 + 4 + length * 4
          when OP_LOOKUPSWITCH
            index += 1
            pad = index % 4
            pad = pad!=0 ? (4-pad) : 0
            index += pad
            npairs = get_int(index+4)
            return pad + 4 + 4 + npairs * 4*2
          when OP_WIDE
            index += 1
            if @code[index] == OP_IINC
              return 5
            else
              return 3
            end
          else raise "don't know how to work out no_of_operands for opcode #{opcode}"
        end
      else return no
    end
  end

  def get_int(index)
    @code[index]<<24 | @code[index+1]<<16 | @code[index+2]<<8 | @code[index+3]
  end

  UNDEFINED = -1
  UNPREDICTABLE = -2
  RESERVED = -3

  NO_OF_OPERANDS = [
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     1, 2, 1, 2, 2, 1, 1, 1, 1, 1,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 1, 1, 1, 1,
     1, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2,
     2, 2, 2, 2, 2, 2, 2, 2, 2,
     2, 2, 2, 2, 2, 1, UNPREDICTABLE, UNPREDICTABLE,
     0, 0, 0, 0, 0, 0,
     2, 2, 2, 2, 2, 2, 2,
     4, UNDEFINED, 2, 1, 2,
     0, 0, 2, 2, 0,
     0, UNPREDICTABLE, 3, 2, 2, 4,
     4, 0, UNDEFINED,
     UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED,
     UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED,
     UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED,
     UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED,
     UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED,
     UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED,
     UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED,
     UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED, UNDEFINED,
     UNDEFINED, UNDEFINED, RESERVED, RESERVED
   ]
end

class LineNumberEntry
  attr_accessor :location, :line_number

  def initialize(location, line_number)
    @location = location
    @line_number = line_number
  end
end

class LineNumberingInfo
  def initialize
    @entries = Array.new
  end

  def add(loc, line_no)
    @entries << LineNumberEntry.new(loc, line_no)
  end

  def each_entry
    @entries.each do |entry|
      yield entry.location, entry.line_number
    end
  end
end

class SourceCoverageInfo
  STATE_NOT_BYTECODE = 0
  STATE_NOT_COVERED = 1
  STATE_COVERED_OK = 2

  def initialize(classes)
    @lines = Array.new
    classes.each do |jclass|
      jclass.each_method do |jmethod|
        coverage = jmethod.coverage
	last_location = last_line_number = nil
        jmethod.line_info.each_entry do |location, line_number|
          unless last_location.nil?
	    coverage_span(coverage, last_location, location-1, last_line_number)
	  end
	  last_location = location
	  last_line_number = line_number
        end
	coverage_span(coverage, last_location, coverage.size-1, last_line_number)
      end
    end
  end

  def coverage_state_of_line(line)
    state = @lines[line]
    return STATE_NOT_BYTECODE if state.nil?

    state
  end

 private

  def coverage_span(coverage, start_location, end_location, line_number)
    start_location.upto(end_location) do |loc|
      if coverage.covered?(loc)
        mark_line(line_number, STATE_COVERED_OK)
	return
      end
    end
    mark_line(line_number, STATE_NOT_COVERED)
  end

  def mark_line(line, state)
    @lines[line] = state
  end
end


