#
# htmlreport.rb
# $Revision: 1.4 $
#
# Copyright (c) 2004 David Holroyd
# See the file COPYING for license terms
#
# HTML report generating plugin for cvprof_report.rb
#
# http://www.badgers-in-foil.co.uk/projects/cvprof/
#

def html_encode(text)
  text.gsub(/./) do |c|
    case c
      when '&' then "&amp;"
      when '<' then "&lt;"
      when '>' then "&gt;"
      else c
    end
  end
end


class HtmlReport
  def initialize
    @base_dir = "html_cvprof_report"
    @current_dir = nil
    @stylesheet_name = "stylesheet.css"
  end

  def add_args(args)
    args << ["--html-dir", "-d", GetoptLong::REQUIRED_ARGUMENT]
  end

  def process_args(opts)
    opts.each do |opt, arg|
      case opt
        when "--html-dir" then @base_dir = arg
      end
    end
  end

  def generate(packages)
    make_dir(@base_dir)
    default_stylesheet
    help
    create("index") do |io|
      head(io, "Java cvprof Coverage Report")
      overview_nav(io)
      io.puts("    <h1>Java cvprof Coverage Report</h1>")
      io.puts("    <table class=\"index\" border=\"1\">")
      io.puts("      <tr><th colspan=\"3\"><h2>Packages</h2></th></tr>")
      packages.each do |jpackage|
        generate_package(jpackage)
        io.puts("      <tr><td><a class=\"package\" href=\"#{package_source_index(jpackage)}\">#{jpackage.name}</a></td></tr>")
      end
      io.puts("    </table>")
      io.puts("  <br />")
      overview_nav(io)
      foot(io)
    end
  end

  def generate_package(jpackage)
    in_dir(package_dir(jpackage)) do
      create("source_index") do |io|
        head(io, "Java cvprof Coverage Report : Sources for Package #{jpackage.name}")
	package_sources_nav(io)
        io.puts("    <h1>Coverage for sources of package<br />#{jpackage.name}</h1>")
        io.puts("    <table class=\"index\" border=\"1\">")
        io.puts("      <tr><th><h2>Source Files</h2></th></tr>")
        jpackage.each_source do |jsource|
          io.puts("      <tr><td><a class=\"source\" href=\"#{jsource.name}.html\">#{jsource.name}</a></td></tr>")
          generate_source(jsource)
        end
        io.puts("    </table>")
        io.puts("  <br />")
	package_sources_nav(io)
        foot(io)
      end
      create("class_index") do |io|
        head(io, "Java cvprof Coverage Report : Classes in Package<br />#{jpackage.name}")
	package_classes_nav(io)
        io.puts("    <h1>Coverage for classes in package #{jpackage.name}</h1>")
        io.puts("    <table class=\"index\" border=\"1\">")
        io.puts("      <tr><th><h2>Classes</h2></th></tr>")
        jpackage.each_class do |jclass|
          io.puts("      <tr><td><a class=\"class\" href=\"#{report_name(jclass)}\">#{jclass.short_name}</a></td></tr>")
#          generate_bytecode(jclass)
        end
        io.puts("    </table>")
        io.puts("  <br />")
	package_classes_nav(io)
        foot(io)
      end
    end
  end

  def generate_source(jsource)
    create(jsource.name) do |io|
      head(io, "Java cvprof Coverage Report : File #{jsource.name}")
      source_nav(io, jsource)
      io.puts("    <h1>Java cvprof Coverage Report for<br />#{jsource.name}</h1>")
      highlight_source_file(io, jsource.path, jsource.create_coverage_info)
      source_nav(io, jsource)
      foot(io)
    end
  end

  HIGHLIGHT_STYLES = ['notbytecode', 'notcovered', 'coveredok', 'partiallycovered']
  STATE_BEGIN = -1

  def highlight_source_file(io, src, coverage)
    File.open(src) do |source_code|
      line_no = 0
      highlight_state = STATE_BEGIN
      io.puts "    <table cellpadding='0' cellspacing='0'>"
      source_code.each_line do |line|
        line_no += 1

        next_state = coverage.coverage_state_of_line(line_no)
        if highlight_state != next_state
          io.puts "</pre></td></tr>" unless highlight_state == STATE_BEGIN
          highlight_state = next_state

          io.print("      <tr><td class=\"#{HIGHLIGHT_STYLES[highlight_state]}\"><pre>")
        end
        io.print "<a name=\"line#{line_no}\" /><span class=\"margin\">"
        io.print sprintf("%5d ", line_no)
        io.print(highlight_state==SourceCoverageInfo::STATE_NOT_COVERED ? "&gt; " : "  ")
        io.print "</span>"
        io.print html_encode(line)
      end
      io.puts "</pre></td></tr>" unless highlight_state == STATE_BEGIN
      io.puts "    </table>"
      io.puts "    <br />"
    end
  end

  def generate_bytecode(jclass)
    create("#{jclass.short_name}-bytecode") do |io|
      head(io, "Java cvprof Coverage Report : Bytecode of class #{jclass.short_name}")
      jclass.each_method do |jmethod|
        io.puts("    <h2>#{jmethod.name_sig}</h2>")
        io.puts "    <table cellpadding='0' cellspacing='0'>"
        highlight_state = STATE_BEGIN
        coverage = jmethod.coverage
        jmethod.bytecode.each_bytecode do |offset, name|
          covered = coverage.covered?(offset)
          next_state = covered ? SourceCoverageInfo::STATE_COVERED_OK : SourceCoverageInfo::STATE_NOT_COVERED
          if highlight_state != next_state
            io.puts "</pre></td></tr>" unless highlight_state == STATE_BEGIN
            highlight_state = next_state

            io.print("      <tr><td class='#{HIGHLIGHT_STYLES[highlight_state]}'><pre>")
          end
          io.print "<span class='margin'>"
          io.printf("%5d", offset)
          io.print(covered ? "  " : "&gt; ")
          io.print "</span>"
          io.puts(html_encode(name))
        end
        io.puts "</pre></td></tr>" unless highlight_state == STATE_BEGIN
        io.puts "    </table>"
      end
      foot(io)
    end
  end

  def report_name(jclass)
    jclass.source + ".html"
  end

  def package_dir(jpackage)
    jpackage.name.gsub(/\./, "/")
  end

  def package_class_index(jpackage)
    package_dir(jpackage) + "/class_index.html"
  end

  def package_source_index(jpackage)
    package_dir(jpackage) + "/source_index.html"
  end

  def source_nav(io, jsource)
    io.puts("<div class=\"mainnav\">")
    io.puts("<a href=\"#{rel_path_to_base_file('index.html')}\">Overview</a>")
    io.puts("<a href=\"source_index.html\">Package Sources</a>")
    io.puts("<a href=\"class_index.html\">Package Classes</a>")
    io.puts("<span class=\"nav_current\">Source File</span>")
    io.puts("<a href=\"#{rel_path_to_base_file('help.html')}\">Help</a>")
    io.puts("</div>")
  end

  def package_sources_nav(io)
    io.puts("<div class=\"mainnav\">")
    io.puts("<a href=\"#{rel_path_to_base_file('index.html')}\">Overview</a>")
    io.puts("<span class=\"nav_current\">Package Sources</span>")
    io.puts("<a href=\"class_index.html\">Package Classes</a>")
    io.puts("<span>Source File</span>")
    io.puts("<a href=\"#{rel_path_to_base_file('help.html')}\">Help</a>")
    io.puts("</div>")
  end

  def package_classes_nav(io)
    io.puts("<div class=\"mainnav\">")
    io.puts("<a href=\"#{rel_path_to_base_file('index.html')}\">Overview</a>")
    io.puts("<a href=\"source_index.html\">Package Sources</a>")
    io.puts("<span class=\"nav_current\">Package Classes</span>")
    io.puts("<span>Source File</span>")
    io.puts("<a href=\"#{rel_path_to_base_file('help.html')}\">Help</a>")
    io.puts("</div>")
  end

  def overview_nav(io)
    io.puts("<div class=\"mainnav\">")
    io.puts("<span class=\"nav_current\">Overview</span>")
    io.puts("<span>Package Sources</span>")
    io.puts("<span>Package Classes</span>")
    io.puts("<span>Source File</span>")
    io.puts("<a href=\"help.html\">Help</a>")
    io.puts("</div>")
  end

  def help_nav(io)
    io.puts("<div class=\"mainnav\">")
    io.puts("<a href=\"index.html\">Overview</a>")
    io.puts("<span>Package Sources</span>")
    io.puts("<span>Package Classes</span>")
    io.puts("<span>Source File</span>")
    io.puts("<span class=\"nav_current\">Help</span>")
    io.puts("</div>")
  end

  def create(filestub)
    File.open("#{pwd}/#{filestub}.html", File::CREAT|File::TRUNC|File::RDWR) do |io|
      yield io
    end
  end

  def head(io, title)
    io.puts "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">"
    io.puts "<html>"
    io.puts "  <head>"
    io.puts "    <title>#{html_encode(title)}</title>"
    io.puts "    <meta name=\"generator\" content=\"cvprof_report #{$version}\" />"
    io.puts "    <link rel=\"stylesheet\" type=\"text/css\" href=\"#{rel_path_to_base_file(@stylesheet_name)}\" />"
    io.puts "  </head>"
    io.puts "  <body>"
  end

  def foot(io)
    io.puts "    <p class=\"footer\"><a href=\"http://www.badgers-in-foil.co.uk/projects/cvprof/\">cvprof</a> #{$version}</p>"
    io.puts "  </body>"
    io.puts "</html>"
  end

  def default_stylesheet
    return unless @stylesheet_name == "stylesheet.css"
    unless FileTest.exists?("#{@base_dir}/stylesheet.css")
      File.open("#{@base_dir}/stylesheet.css", File::CREAT|File::TRUNC|File::RDWR) do |io|
        io.print <<SHEET_END
/*
 * Default cascading stylesheet for cvprof report
 *
 * If this file exists, the report generator won't overwrite it, so you should
 * be safe to edit this and later re-run the report
 */

h1, h2, h3 {
	font-family: sans-serif;
}

/* summary/nav styles */
table.index {
	width: 100%;
}
th {
	background: #EEEEFF;
}

a.package, a.class {
	font-weight: bolder;
}

/* Source code styling */
span.margin {
	color:#888888; background-color:#eeeeee;
}
.notcovered {
	background-color:#ffdddd;
}
.partiallycovered {
	background-color:#ffeedd;
}
.notbytecode {
	color:#aaaaaa;
}

.footer {
	text-align: center;
	font-size: smaller;
}
.mainnav {
	background-color: #EEEEFF;
	padding: 4px;
}
.mainnav > * {
	font-family: sans-serif;
	font-weight: bolder;
	padding: 4px;
}
.nav_current {
	background-color: #00008B;
	color: #FFFFFF;
}
/* stop source code sections getting spaced out */
pre {
	margin:0px;
}

SHEET_END
      end
    end
  end

  def help
    create("help") do |io|
      head(io, "Java cvprof Coverage Report : Help")
      help_nav(io)
      io.print <<HELP_END
    <a name="overview" />
    <h1>Overview Page</h1>
    <p><i>[todo]</i></p>

    <a name="package_sources" />
    <h1>Package Sources Pages</h1>
    <p><i>[todo]</i></p>


    <a name="package_classes" />
    <h1>Package Classes Pages</h1>
    <p><i>[todo]</i></p>


    <a name="source_file" />
    <h1>Source File Pages</h1>

    <p>Lines that are shaded less distinctly than others are 'not executable'
    and didn't contribute any bytecode.</p>
    <table>
      <tr><td class="notbytecode"><pre><span class="margin">    8   </span>	public void test() {</pre></td></tr>
    </table>

    <p>Lines with more contrast are 'executable', and some of their bytecode
    was executed during profiling.</p>
    <table>
      <tr><td class="coveredok"><pre><span class="margin">    9   </span>		int temp = a;</pre></td></tr>
    </table>

    <p>When a chevron appears in the margin, and the background of the line
    is coloured red, none of its bytecode was executed during profiling.</p>
    <table>
      <tr><td class="notcovered"><pre><span class="margin">   39 &gt; </span>		return 0;</pre></td></tr>
    </table>

    <br />
HELP_END
      help_nav(io)
      foot(io)
    end
  end

  def make_dir(name)
    path = nil
    name.split(/\//).each do |part|
      if path.nil?
        path = part
      else
        path = "#{path}/#{part}"
      end
      unless FileTest.directory?(path)
        Dir.mkdir(path)
      end
    end
  end

  def in_dir(name)
    old_dir = @current_dir
    @current_dir = name
    make_dir(pwd)
    yield
    @current_dir = old_dir
  end

  def pwd
    if @current_dir.nil?
      @base_dir
    else
      "#{@base_dir}/#{@current_dir}"
    end
  end

  def rel_path_to_base_file(name)
    return name if @current_dir.nil?
    path = "../"
    @current_dir.scan(/\//) do
      path << "../"
    end
    path + name
  end
end
