/*
 * bitset.h
 * $Revision: 1.2 $
 *
 * Copyright (c) 2004 David Holroyd
 * See the file COPYING for license terms
 *
 * http://www.badgers-in-foil.co.uk/projects/cvprof/
 *
 */


/*
 * A bitset for recording which bytecode locations have been covered
 */


typedef unsigned long bitset_t;

typedef struct {
	/* number of bits in bitset */
	int size;
	bitset_t bytecodes[0];
} coverage;


static inline int cov_slots(bits)
{
	return (bits-1) / (sizeof(bitset_t)*8) + 1;
}

static inline int cov_size(int bits)
{
	return cov_slots(bits) * sizeof(bitset_t);
}

coverage *coverage_allocate(int bytecode_size)
{
	coverage *result;
	int size = cov_size(bytecode_size);
	result = (coverage *)malloc(sizeof(coverage) + size);
	result->size = bytecode_size;
	memset(result->bytecodes, 0, size);
	return result;
}

void coverage_free(coverage *cov)
{
	free(cov);
}

int coverage_sizeof(coverage *cov)
{
	return sizeof(coverage) + cov_size(cov->size);
}

int coverage_bit_count(coverage *cov)
{
	return cov->size;
}

static inline int out_of_bounds(int size, int bytecode)
{
	if (bytecode < 0 || bytecode >= size) {
		fprintf(stderr, "COVERAGE: bytecode index outside method (0 to %d): %d\n",
		        (size - 1), bytecode);
		return 1;
	}
	return 0;
}

inline void coverage_mark(coverage *cov, int bytecode)
{
	int slot = bytecode / (sizeof(bitset_t)*8);
	int element = bytecode - sizeof(bitset_t)*8*slot;

	if (out_of_bounds(cov->size, bytecode)) return;

	cov->bytecodes[slot] |= (1UL<<element);
}

inline int coverage_is_marked(coverage *cov, int bytecode)
{
	int slot = bytecode / (sizeof(bitset_t)*8);
	int element = bytecode - sizeof(bitset_t)*8*slot;

	if (out_of_bounds(cov->size, bytecode)) return 0;

	return (cov->bytecodes[slot] & (1UL<<element)) != 0;
}

/* for debugging */
void coverage_print(FILE *f, coverage *cov)
{
	int i = cov->size;
	fprintf(f, "cov[%d]:", i);

	while (--i >= 0) {
		if (coverage_is_marked(cov, i)) {
			fprintf(f, "1");
		} else {
			fprintf(f, "0");
		}
	}
	fprintf(f, "\n");
}
