/*
 * cvprof.c
 * $Revision: 1.6 $
 *
 * Copyright (c) 2004 David Holroyd
 * See the file COPYING for license terms
 *
 * http://www.badgers-in-foil.co.uk/projects/cvprof/
 */

#include <stdlib.h>
#include <string.h>

/*
 * jni_md.h assumes visual studio '__int64' will be available under windows.
 * typedef it to gcc equivalent, 'long long'
 */
#ifndef __int64
#  ifdef __GNUC__
     typedef long long __int64;
#  endif
#endif

#define NO_JVMDI_MACROS
#include <jvmdi.h>
#include <jni.h>

#include "bitset.h"
#include "hash.h"
#include "exclusions.h"
#include "options.h"
#include "err.h"

#define PROFILER_UNINITIALISED 0
#define PROFILER_INITIALISED 1
#define PROFILER_SHUTDOWN -1

/* for SetEventNotificationMode's 'thread' argument */
#define ALL_THREADS NULL

#define DEFAULT_DB "coverage"

#define ACC_ABSTRACT 0x0400

typedef struct {
	int line_count;
	JVMDI_line_number_entry lines[0];
} lineinfo;


typedef struct {
	int bytecode_count;
	jbyte bytecodes[0];
} bytecodeinfo;


/* Interface to the Java debug sub-system */
JVMDI_Interface_1 *jvmdi;

/* Flag to determine initialisation has occurred */
int init = PROFILER_UNINITIALISED;

/* Used to control muti-threaded access to profile data */
JVMDI_RawMonitor profile_lock;


exclusion_list *exclude_packages = NULL;

/* options -- constructed from command line args */
options opts;

/* Handle on the profile data itself */
hash_table hash;


/*
 * TODO
 *  - compile this with gprof instrumentation
 *  - probably cache the hash datums, and only write if thisMethod != lastMethod
 *  - look at alternatives to dbm (mmap?)
 *  - record bytecode data so that we can tell which positions in a coverage
 *    bitmap do not actualy mark the position of a bytecode
 */

static void report_jvmdi_error(char *message, jvmdiError err)
{
	char *error_text = NULL;
	switch (err) {
	    case JVMDI_ERROR_INVALID_THREAD:
		error_text = "JVMDI_ERROR_INVALID_THREAD";
		break;
	    case JVMDI_ERROR_INVALID_THREAD_GROUP:
		error_text = "JVMDI_ERROR_INVALID_THREAD_GROUP";
		break;
	    case JVMDI_ERROR_INVALID_PRIORITY:
		error_text = "JVMDI_ERROR_INVALID_PRIORITY";
		break;
	    case JVMDI_ERROR_THREAD_NOT_SUSPENDED:
		error_text = "JVMDI_ERROR_THREAD_NOT_SUSPENDED";
		break;
	    case JVMDI_ERROR_THREAD_SUSPENDED:
		error_text = "JVMDI_ERROR_THREAD_SUSPENDED";
		break;
	    case JVMDI_ERROR_INVALID_OBJECT:
		error_text = "JVMDI_ERROR_INVALID_OBJECT";
		break;
	    case JVMDI_ERROR_INVALID_CLASS:
		error_text = "JVMDI_ERROR_INVALID_CLASS";
		break;
	    case JVMDI_ERROR_CLASS_NOT_PREPARED:
		error_text = "JVMDI_ERROR_CLASS_NOT_PREPARED";
		break;
	    case JVMDI_ERROR_INVALID_METHODID:
		error_text = "JVMDI_ERROR_INVALID_METHODID";
		break;
	    case JVMDI_ERROR_INVALID_LOCATION:
		error_text = "JVMDI_ERROR_INVALID_LOCATION";
		break;
	    case JVMDI_ERROR_INVALID_FIELDID:
		error_text = "JVMDI_ERROR_INVALID_FIELDID";
		break;
	    case JVMDI_ERROR_INVALID_FRAMEID:
		error_text = "JVMDI_ERROR_INVALID_FRAMEID";
		break;
	    case JVMDI_ERROR_NO_MORE_FRAMES:
		error_text = "JVMDI_ERROR_NO_MORE_FRAMES";
		break;
	    case JVMDI_ERROR_OPAQUE_FRAME:
		error_text = "JVMDI_ERROR_OPAQUE_FRAME";
		break;
	    case JVMDI_ERROR_NOT_CURRENT_FRAME:
		error_text = "JVMDI_ERROR_NOT_CURRENT_FRAME";
		break;
	    case JVMDI_ERROR_TYPE_MISMATCH:
		error_text = "JVMDI_ERROR_TYPE_MISMATCH";
		break;
	    case JVMDI_ERROR_INVALID_SLOT:
		error_text = "JVMDI_ERROR_INVALID_SLOT";
		break;
	    case JVMDI_ERROR_DUPLICATE:
		error_text = "JVMDI_ERROR_DUPLICATE";
		break;
	    case JVMDI_ERROR_NOT_FOUND:
		error_text = "JVMDI_ERROR_NOT_FOUND";
		break;
	    case JVMDI_ERROR_INVALID_MONITOR:
		error_text = "JVMDI_ERROR_INVALID_MONITOR";
		break;
	    case JVMDI_ERROR_NOT_MONITOR_OWNER:
		error_text = "JVMDI_ERROR_NOT_MONITOR_OWNER";
		break;
	    case JVMDI_ERROR_INTERRUPT:
		error_text = "JVMDI_ERROR_INTERRUPT";
		break;
	    case JVMDI_ERROR_INVALID_CLASS_FORMAT:
		error_text = "JVMDI_ERROR_INVALID_CLASS_FORMAT";
		break;
	    case JVMDI_ERROR_CIRCULAR_CLASS_DEFINITION:
		error_text = "JVMDI_ERROR_CIRCULAR_CLASS_DEFINITION";
		break;
	    case JVMDI_ERROR_FAILS_VERIFICATION:
		error_text = "JVMDI_ERROR_FAILS_VERIFICATION";
		break;
	    case JVMDI_ERROR_ADD_METHOD_NOT_IMPLEMENTED:
		error_text = "JVMDI_ERROR_ADD_METHOD_NOT_IMPLEMENTED";
		break;
	    case JVMDI_ERROR_SCHEMA_CHANGE_NOT_IMPLEMENTED:
		error_text = "JVMDI_ERROR_SCHEMA_CHANGE_NOT_IMPLEMENTED";
		break;
	    case JVMDI_ERROR_INVALID_TYPESTATE:
		error_text = "JVMDI_ERROR_INVALID_TYPESTATE";
		break;
	    case JVMDI_ERROR_NOT_IMPLEMENTED:
		error_text = "JVMDI_ERROR_NOT_IMPLEMENTED";
		break;
	    case JVMDI_ERROR_NULL_POINTER:
		error_text = "JVMDI_ERROR_NULL_POINTER";
		break;
	    case JVMDI_ERROR_ABSENT_INFORMATION:
		error_text = "JVMDI_ERROR_ABSENT_INFORMATION";
		break;
	    case JVMDI_ERROR_INVALID_EVENT_TYPE:
		error_text = "JVMDI_ERROR_INVALID_EVENT_TYPE";
		break;
	    case JVMDI_ERROR_ILLEGAL_ARGUMENT:
		error_text = "JVMDI_ERROR_ILLEGAL_ARGUMENT";
		break;
	    case JVMDI_ERROR_OUT_OF_MEMORY:
		error_text = "JVMDI_ERROR_OUT_OF_MEMORY";
		break;
	    case JVMDI_ERROR_ACCESS_DENIED:
		error_text = "JVMDI_ERROR_ACCESS_DENIED";
		break;
	    case JVMDI_ERROR_VM_DEAD:
		error_text = "JVMDI_ERROR_VM_DEAD";
		break;
	    case JVMDI_ERROR_INTERNAL:
		error_text = "JVMDI_ERROR_INTERNAL";
		break;
	    case JVMDI_ERROR_UNATTACHED_THREAD:
		error_text = "JVMDI_ERROR_UNATTACHED_THREAD";
		break;
	}
	if (error_text == NULL) {
		EPRINT("%s: JVMDI error code %d", message, (int)err);
	} else {
		EPRINT("%s: %s", message, error_text);
	}
}

static inline void profile_lock_enter()
{
	jvmdiError err;
	err = jvmdi->RawMonitorEnter(profile_lock);
	if (err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't aquire monitor", err);
		exit(-2);
	}
}


static inline void profile_lock_exit()
{
	jvmdiError err;
	err = jvmdi->RawMonitorExit(profile_lock);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't release monitor", err);
	}
}


/*
 * Returns 0 on success
 */
static int setup_package_exclusion_list()
{
	if (opts.exclusions_file == NULL) {
		exclusion_list_add(&exclude_packages, strdup("Lsun/"));
		exclusion_list_add(&exclude_packages, strdup("Ljava/"));
		exclusion_list_add(&exclude_packages, strdup("Ljavax/"));
		exclusion_list_add(&exclude_packages,strdup("Lcom/sun/"));
		return 0;
	} else {
		FILE *in = fopen(opts.exclusions_file, "r");
		if (!in) {
			EPRINT("Can't open exclusions file \"%s\"",
			       opts.exclusions_file);
			return 1;
		}
		exclude_packages = exclusion_list_read(in);
		fclose(in);

		return 0;
	}
}


static int should_profile_class(jclass clazz)
{
	jvmdiError err;
	jboolean is_interface;
	char *clazz_sig;
	int result = 1;

	err = jvmdi->IsInterface(clazz, &is_interface);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't determine if class is an interface", err);
		return 0;
	}

	if (is_interface) {
		return 0;
	}

	err = jvmdi->GetClassSignature(clazz, &clazz_sig);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't get class details", err);
		return 0;
	}

	if (exclusion_list_matches(exclude_packages, clazz_sig)) {
		result = 0;
	}
	jvmdi->Deallocate((jbyte *)clazz_sig);

	return result;
}


/*
 * The resulting string should be free()d when the calling code is finished
 * with it.
 */
static char *make_class_key(JNIEnv *env, jclass clazz, const char *suffix)
{
	jvmdiError err;
	char *clazz_sig = NULL;
	char *key_cursor;
	char *result;


	err = jvmdi->GetClassSignature(clazz, &clazz_sig);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't get class details", err);
		return NULL;
	}

	result = malloc(strlen(clazz_sig) + strlen(suffix) + 1);

	/*
	 * REVISIT: jvmdicover can include the hashCode of the classloder in
	 *          the key name.  Here's the place to add similar code to this
	 *          software.
	 */

	if (result) {
		key_cursor = result;
		strcpy(key_cursor, clazz_sig);
		key_cursor += strlen(clazz_sig);
		strcpy(key_cursor, suffix);
	} else {
		EPRINT("make_class_key() failed to allocate mem for result");
	}

	jvmdi->Deallocate((jbyte *)clazz_sig);


	return result;
}


/*
 * The resulting string should be free()d when the calling code is finished
 * with it.
 */
static char *make_method_key(JNIEnv *env, jclass clazz, jmethodID method,
                            const char *suffix)
{
	jvmdiError err;
	char *method_name = NULL;
	char *method_sig = NULL;
	char *method_suffix;
	char *result;

	err = jvmdi->GetMethodName(clazz, method, &method_name, &method_sig);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't get method details", err);
		return NULL;
	}

	method_suffix = malloc(strlen(method_name) + strlen(method_sig)
	                        + strlen(suffix) + 1);

	if (method_suffix) {
		char *key_cursor = method_suffix;

		strcpy(key_cursor, method_name);
		key_cursor += strlen(method_name);

		strcpy(key_cursor, method_sig);
		key_cursor += strlen(method_sig);

		strcpy(key_cursor, suffix);
	}
else EPRINT("make_method_key() : failed to allocate mem for method_suffix");

	jvmdi->Deallocate((jbyte *)method_name);
	jvmdi->Deallocate((jbyte *)method_sig);

	if (!method_suffix) return NULL;


	result = make_class_key(env, clazz, method_suffix);

	free(method_suffix);

	return result;
}


/*
 * Returns 0 on success
 */
static int coverage_put(JNIEnv *env, jclass clazz, jmethodID method,
                         coverage *cov)
{
	int result;
	char *key = make_method_key(env, clazz, method, ":coverage");

	if (!key) return 1;

	result = hash_put(hash, key, coverage_sizeof(cov), (void *)cov);

	free(key);

	return result;
}


/*
 * Returns the coverage bitmap for the given method, or null if there is a
 * problem, or the method isn't being profiled.
 * 
 * The returned structure should not be freed.
 */
static coverage *coverage_get(JNIEnv *env, jclass clazz, jmethodID method)
{
	coverage *result;
	int size;
	char *key;

	key = make_method_key(env, clazz, method, ":coverage");

	if (!key) {
		return NULL;
	}

	if (hash_get(hash, key, &size, (void **)&result)) {
		result = NULL;
	}

	free(key);

	return result;
}

/*
jint last_clazz = NULL;
jmethodID last_method = NULL;
*/

static void update_coverage(JNIEnv *env, JVMDI_single_step_event_data *event)
{
	jclass clazz = event->clazz;
	jmethodID method = event->method;
/*	jvmdiError err;
	jint clazz_code;

	err = jvmdi->GetObjectHashCode(clazz, &clazz_code);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("GetObjectHashCode failed", err);
	}

	if (last_clazz==clazz_code && last_method==method) {
		EPRINT("HIT");
	} else {
		EPRINT("miss");
	}

	last_clazz = clazz_code;
	last_method = method;
*/	
	coverage *cov = coverage_get(env, clazz, method);

	if (cov) {
		coverage_mark(cov, event->location);
		coverage_put(env, clazz, method, cov);
	}
}


static int should_profile_method(jclass clazz, jmethodID method)
{
	jvmdiError err;
	jboolean method_type;
	jint modifiers;

	err = jvmdi->GetMethodModifiers(clazz, method, &modifiers);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("GetMethodModifiers failed", err);
		return 0;
	}
	/* TODO: do we need to check for 'native' too? */
	if (modifiers & ACC_ABSTRACT) {
		return 0;
	}

	err = jvmdi->IsMethodSynthetic(clazz, method, &method_type);
	if(err != JVMDI_ERROR_NONE) {
		if (err != JVMDI_ERROR_NOT_IMPLEMENTED) {
			report_jvmdi_error("IsMethodSynthetic failed", err);
		}
		return 0;
	}
	if (method_type) return 0;

	err = jvmdi->IsMethodNative(clazz, method, &method_type);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("IsMethodNative failed", err);
		return 0;
	}
	if (method_type) return 0;

	return 1;
}


static int should_profile_thread(JNIEnv *env, jthread thread)
{
	/*
	 * TODO: test for threads (probably by name) that we don't need to
	 *       enable single-stepping/profiling for, e.g. "Signal Dispatcher",
	 *       "process reaper", "SIGINT handler"
	 */
	return 1;
}


static void record_source_file_name(JNIEnv *env, jclass clazz,
                                    const char *source_file)
{
	char *key = make_class_key(env, clazz, ":source");

	if (key) {
		hash_put(hash, key, strlen(source_file), source_file);
		free(key);
	}
}


static void create_line_number_data(JNIEnv *env, jclass clazz, jmethodID method)
{
	jvmdiError err;
	jint line_number_count;
	JVMDI_line_number_entry *line_number_table;

	int table_size;
	lineinfo *table_copy;
	char *key;

	err = jvmdi->GetLineNumberTable(clazz, method,
	                                &line_number_count,
	                                &line_number_table);
	if (err != JVMDI_ERROR_NONE) {
		if (err != JVMDI_ERROR_ABSENT_INFORMATION) {
			report_jvmdi_error("Error retriving line information",
			                   err);
		}
		/* else, Maybe native method or no debug. Ignore */
		return;
	}

	key = make_method_key(env, clazz, method, ":lines");
	if (key) {
		table_size = sizeof(lineinfo) + sizeof(JVMDI_line_number_entry)
		            * line_number_count;

		table_copy = (lineinfo *)malloc(table_size);
		if (table_copy) {
			table_copy->line_count = line_number_count;
			memcpy(table_copy->lines,
			       line_number_table,
			   sizeof(JVMDI_line_number_entry) * line_number_count);

			if (hash_put(hash, key, table_size, table_copy)){
				EPRINT("Writing line number data to db failed");
			}
			free(table_copy);
		}
		free(key);
	}

	jvmdi->Deallocate((jbyte *)line_number_table);
}


/*
 * Returns the length of this method, or -1 if the method length can't be
 * determined.  Note this will usually be greater than the number of
 * bytecodes, since the length includes both bytecodes and 'immediate operands'.
 *
 * If the method is abstract (from an abstract class, or from an interface)
 * the result is undefined.
 */
static int get_method_length(jclass clazz, jmethodID method)
{
	jvmdiError err;
	jlocation start;
	jlocation end;

	err = jvmdi->GetMethodLocation(clazz, method, &start, &end);

	if (err != JVMDI_ERROR_NONE) {
		if (err != JVMDI_ERROR_ABSENT_INFORMATION) {
			report_jvmdi_error("Error getting method location information",
			                   err);
		}
		return -1;
	}

	if (start == -1 || end == -1) {
		return -1;
	}

	return end - start + 1;
}

static void create_bytecode_data(JNIEnv *env, jclass clazz,
                                         jmethodID method)
{
	jvmdiError err;
	jint bytecode_count;
	jbyte *bytecodes;
	char *key;

	err = jvmdi->GetBytecodes(clazz, method,
	                                 &bytecode_count, &bytecodes);
	if (err != JVMDI_ERROR_NONE) {
		if (err != JVMDI_ERROR_NOT_IMPLEMENTED) {
			report_jvmdi_error("Couldn't get bytecodes", err);
		}
		return;
	}

	key = make_method_key(env, clazz, method, ":bytecode");
	if (key) {
		int bytecode_size = sizeof(bytecodeinfo)
		                  + sizeof(jbyte)*bytecode_count;
		bytecodeinfo *value = (bytecodeinfo *)malloc(bytecode_size);
		value->bytecode_count = bytecode_count;
		memcpy(value->bytecodes, bytecodes,
		       sizeof(jbyte) * bytecode_count);

		if (hash_put(hash, key, bytecode_size, value)) {
			EPRINT("Writing line number data to db failed");
		}
	}

	jvmdi->Deallocate(bytecodes);
}


static void create_initial_coverage_data(JNIEnv *env, jclass clazz,
                                         jmethodID method)
{
	coverage *cov;

	int bitset_size = get_method_length(clazz, method);
	cov = coverage_allocate(bitset_size);
	if (!cov) return;

	coverage_put(env, clazz, method, cov);

	coverage_free(cov);
}


static void create_method_data(JNIEnv *env, jclass clazz, jmethodID method)
{
	create_line_number_data(env, clazz, method);

	create_bytecode_data(env, clazz, method);

	create_initial_coverage_data(env, clazz, method);
}


static void create_class_data(JNIEnv *env, jclass clazz)
{
	jvmdiError err;
	char *source_file;
	jmethodID *methods;
	jint methodCount;
	int i;

	err = jvmdi->GetSourceFileName(clazz, &source_file);
	if (err == JVMDI_ERROR_NONE) {
		record_source_file_name(env, clazz, source_file);
		jvmdi->Deallocate((jbyte *)source_file);
	} else if (err != JVMDI_ERROR_ABSENT_INFORMATION) {
		report_jvmdi_error("GetSourceFileName failed", err);
	}

	err = jvmdi->GetClassMethods(clazz, &methodCount, &methods);
	if(err != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Couldn't get class methods", err);
		return;
	}

	for(i = 0 ; i<methodCount ; i++) {
		jmethodID method = methods[i];
		if (should_profile_method(clazz, method)) {
			create_method_data(env, clazz, method);
		}
	}
	jvmdi->Deallocate((jbyte *)methods);
}


static inline void handle_vm_init(JNIEnv *env)
{
	if(jvmdi->CreateRawMonitor("Profile", &profile_lock)!=JVMDI_ERROR_NONE){
		EPRINT("Can't create raw monitor lock");
	}

	init = PROFILER_INITIALISED;
}


static inline void handle_vm_death(JNIEnv *env)
{
	/* try to make sure nothing holds the lock when we destry it */
	init = PROFILER_SHUTDOWN;
	profile_lock_enter();
	profile_lock_exit();

	jvmdi->DestroyRawMonitor(profile_lock);
	hash_close(hash);
	exclusion_list_free(exclude_packages);
	options_free(&opts);
}


static inline void handle_single_step(JNIEnv *env,
                                      JVMDI_single_step_event_data *event)
{
	profile_lock_enter();
	update_coverage(env, event);
	profile_lock_exit();
}


static inline void handle_class_prepare(JNIEnv *env,
                                        JVMDI_class_event_data *event)
{
	jclass clazz = event->clazz;
	if (should_profile_class(clazz)) {
		profile_lock_enter();
		create_class_data(env, clazz);
		profile_lock_exit();
	}
}

static inline void handle_thread_start(JNIEnv *env,
                                       JVMDI_thread_change_event_data *event)
{
	jvmdiError err;

	if (should_profile_thread(env, event->thread)) {
		err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
		                                      JVMDI_EVENT_SINGLE_STEP,
		                                      event->thread);
		if(err != JVMDI_ERROR_NONE)
			report_jvmdi_error("Can't enable single step", err);
	}
}

/*
 * If we don't disable single-step events, the Windows 1.3 JVM hangs at exit,
 * so this handler disables notification for each thread as it ends.  This
 * handle may get called after JVMDI_EVENT_VM_DEATH is recieved, so care must
 * be taken not to touch any resources (e.g. the lock) that that event handler
 * releases.
 */
static inline void handle_thread_end(JNIEnv *env,
                                       JVMDI_thread_change_event_data *event)
{
	jvmdiError err;
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_SINGLE_STEP,
	                                      event->thread);
	if(err != JVMDI_ERROR_NONE)
		report_jvmdi_error("Can't enable single step", err);
}


static inline int should_ignore_event(JVMDI_Event *event)
{
	/* Allow THREAD_END after VM_DEATH so we can clean up */
	if(init!=PROFILER_INITIALISED
	   && event->kind!=JVMDI_EVENT_VM_INIT
	   && event->kind!=JVMDI_EVENT_THREAD_END)
	{
		if(init == PROFILER_UNINITIALISED) {
			EPRINT("Unexpected event before VM init: %d",
			       (int)event->kind);
		}
		/* you can get events after VM_DEATH */
		return 1;
	}
	return 0;
}


JNIEXPORT void hook(JNIEnv *env, JVMDI_Event *event)
{
	if (should_ignore_event(event)) return;

	switch(event->kind) {
	    case JVMDI_EVENT_VM_INIT:
		handle_vm_init(env);
		break;
	    case JVMDI_EVENT_VM_DEATH:
		handle_vm_death(env);
		break;
	    case JVMDI_EVENT_SINGLE_STEP:
		handle_single_step(env, &event->u.single_step);
		break;
	    case JVMDI_EVENT_CLASS_PREPARE:
		handle_class_prepare(env, &event->u.class_event);
		break;
	    case JVMDI_EVENT_THREAD_START:
		handle_thread_start(env, &event->u.thread_change);
		break;
	    case JVMDI_EVENT_THREAD_END:
		handle_thread_end(env, &event->u.thread_change);
		break;
	    default:
		EPRINT("Unexpected event type: %d", (int)event->kind);
	}
}


static void check_jvmdi_caps(JVMDI_capabilities *caps)
{
	if(!caps->can_get_synthetic_attribute) {
		EPRINT("Can't get synthetic attribute; profile will include synthesized methods");
	}
	if(!caps->can_get_bytecodes) {
		EPRINT("Can't get bytecodes; may be a problem once reporting tools need to use them");
	}
}


static jvmdiError enable_jvmdi_events()
{
	jvmdiError err;

	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_VM_INIT,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable VM init event");
	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_VM_DEATH,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable VM death event");

	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_THREAD_START,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable thread start event");
	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_THREAD_END,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable thread end event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_SINGLE_STEP,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		report_jvmdi_error("Can't disable single step event", err);
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_EXCEPTION_CATCH,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable exception catch event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_BREAKPOINT,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable breakpoint event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_FIELD_MODIFICATION,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable field modification event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_FIELD_ACCESS,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable field access event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_FRAME_POP,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable frame pop event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_METHOD_ENTRY,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable method entry event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_METHOD_EXIT,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable method exit event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_EXCEPTION,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable exception event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_USER_DEFINED,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable user defined event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_CLASS_UNLOAD,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable class unload event");
	err = jvmdi->SetEventNotificationMode(JVMDI_DISABLE,
	                                      JVMDI_EVENT_CLASS_LOAD,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't disable class load event");
	err = jvmdi->SetEventNotificationMode(JVMDI_ENABLE,
	                                      JVMDI_EVENT_CLASS_PREPARE,
	                                      ALL_THREADS);
	if(err != JVMDI_ERROR_NONE)
		EPRINT("Can't enable class prepare event");

	return jvmdi->SetEventHook(hook);
}


JNIEXPORT jint JNICALL JVM_OnLoad(JavaVM *jvm, void *options, void *reserved)
{
	int jvmError;
	jvmdiError err;
	JVMDI_capabilities caps;

	jvmError=(*jvm)->GetEnv(jvm, (void **)&jvmdi, JVMDI_VERSION_1);
	if(jvmError == JNI_EDETACHED) {
		EPRINT("VM Detached");
		return JNI_ERR;
	}
	if(jvmError == JNI_EVERSION) {
		EPRINT("Bad JVMDI version");
		return JNI_ERR;
	}

	if (options_process_list(&opts, (char *)options)) {
		return JNI_EINVAL;
	}

	setup_package_exclusion_list();

	err = jvmdi->GetCapabilities(&caps);
	if(err != JVMDI_ERROR_NONE) {
		if(err == JVMDI_ERROR_ACCESS_DENIED) {
			EPRINT("Profile can't be started since debugging is disabled. Specify -Xdebug on JVM command line (SUN JVM)");
			return JNI_ERR;
		}
		report_jvmdi_error("Can't get capabilities", err);
	}
	check_jvmdi_caps(&caps);

	if (enable_jvmdi_events() != JVMDI_ERROR_NONE) {
		report_jvmdi_error("Can't set event hook", err);
		return JNI_ERR;
	}

	hash = hash_open(DEFAULT_DB);
	if (!hash) {
		return JNI_ERR;
	}

	return JNI_OK;
}
